"""
route for API document page and OpenAPI spec
"""
import os
import re


class RedocPage:
    def __init__(self, config):
        self.config = config

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        with open(os.path.join(os.getcwd(),
                               'falibrary',
                               self.config.TEMPLATE_FOLDER,
                               'redoc.html'), 'r', encoding='utf-8') as f:
            page = f.read()

        resp.body = re.sub('{{}}', self.config.SPEC_URL, page)


class OpenAPI:
    def __init__(self, api):
        self.api = api

    def on_get(self, req, resp):
        resp.media = self.api.spec


_doc_class_name = [x.__name__ for x in (RedocPage, OpenAPI)]



import re
from functools import wraps, partial
from pydantic import ValidationError, BaseModel
import falcon

from falibrary.config import default_config
from falibrary.route import OpenAPI, RedocPage
from falibrary.utils import find_routes, parse_path


class Falibrary:
    """
    :param app: Falcon instance
    :param kwargs: key-value for config
    """

    def __init__(self, app=None, **kwargs):
        self.app = app
        self.models = {}
        self.config = default_config
        for key, value in kwargs.items():
            setattr(self.config, key.upper(), value)

        self.STATUS = re.compile(r'(?P<code>^\d{3}) (?P<msg>[\w ]+$)')
        if self.app:
            assert isinstance(app, falcon.API)
            self._register_route()

    def register(self, app):
        assert isinstance(app, falcon.API)
        self.app = app
        self._register_route()

    def validate(self, query=None, data=None, resp=None, x=[]):
        """
        validate query, JSON data, and response according to
        ``pydantic.BaseModel``

        :param query: Schema for query args
        :param data: Schema for JSON data
        :param response: Schema for JSON response
        :param x: List of :class:`falcon.status_codes`
        """
        def decorator_validation(func):
            @wraps(func)
            def validation(self, _req, _resp, *args, **kwargs):
                try:
                    if query:
                        setattr(_req.context, 'query', query(**_req.params))
                    if data:
                        setattr(_req.context, 'data', data(**_req.media))
                except ValidationError as err:
                    raise falcon.HTTPUnprocessableEntity(
                        'Schema failed validation',
                        description=str(err),
                    )
                except Exception:
                    raise

                response = func(self, _req, _resp, *args, **kwargs)
                if resp:
                    _resp.media = response.dict()

                return response

            # register ``pydantic.BaseModel``
            for name, model in zip(
                ('query', 'data', 'resp'), (query, data, resp)
            ):
                if model:
                    assert issubclass(model, BaseModel)
                    self.models[model.__name__] = model.schema()
                    setattr(validation, name, model.__name__)

            # handle exceptions
            code_msg = {}
            for exception in x:
                match = self.STATUS.match(exception)
                assert match
                code_msg[match.group('code')] = match.group('msg')

            if code_msg:
                validation.x = code_msg

            return validation
        return decorator_validation

    def _register_route(self):
        """
        register doc page and OpenAPI spec file
        """
        self.config.SPEC_URL = f'/{self.config.PATH}/{self.config.FILENAME}'
        self.app.add_route(
            f'/{self.config.PATH}',
            RedocPage(self.config)
        )
        self.app.add_route(
            self.config.SPEC_URL,
            OpenAPI(self)
        )

    @property
    def spec(self):
        if not hasattr(self, '_spec'):
            self._generate_spec()
        return self._spec

    def _generate_spec(self):
        routes = {}
        for route in find_routes(self.app._router._roots):
            path, parameters = parse_path(route.uri_template)
            routes[path] = {}
            for method, func in route.method_map.items():
                if isinstance(func, partial):
                    # ignore exception handlers
                    continue

                name = route.resource.__class__.__name__
                spec = {
                    'summary': name,
                    'operationID': name + '__' + method.lower(),
                }

                if hasattr(func, 'query'):
                    spec['requestBody'] = {
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': f'#/components/schemas/{func.data}'
                                }
                            }
                        }
                    }

                if hasattr(func, 'query'):
                    parameters.append({
                        'name': func.query,
                        'in': 'query',
                        'required': True,
                        'schema': {
                            '$ref': f'#/components/schemas/{func.query}',
                        }
                    })
                spec['parameters'] = parameters

                if hasattr(func, 'resp'):
                    spec['responses'] = {
                        '200': {
                            'description': 'Successful Response',
                            'content': {
                                'application/json': {
                                    'schema': {
                                        '$ref': f'#/components/schemas/{func.resp}'
                                    }
                                }
                            }
                        }
                    }
                else:
                    spec['responses'] = {
                        '200': {
                            'description': 'Successful Response',
                        }
                    }

                if any([hasattr(func, schema)
                        for schema in ('query', 'data', 'resp')]):
                    spec['responses']['422'] = {
                        'description': 'Validation Error',
                    }

                if hasattr(func, 'x'):
                    for code, msg in func.x.items():
                        spec['responses'][str(code)] = {
                            'description': msg,
                        }

                routes[path][method.lower()] = spec

        definitions = {}
        for _, schema in self.models.items():
            if 'definitions' in schema:
                for key, value in schema['definitions'].items():
                    definitions[key] = value
                del schema['definitions']

        data = {
            'openapi': self.config.OPENAPI_VERSION,
            'info': {
                'title': self.config.TITLE,
                'version': self.config.VERSION,
            },
            'paths': {
                **routes
            },
            'components': {
                'schemas': {
                    name: schema for name, schema in self.models.items()
                },
            },
            'definitions': definitions
        }
        self._spec = data



"""
utils functions
"""
import re
from falcon.routing.compiled import _FIELD_PATTERN

from falibrary.route import _doc_class_name

# NOTE from `falcon.routing.compiled.CompiledRouterNode`
ESCAPE = r'[\.\(\)\[\]\?\$\*\+\^\|]'
ESCAPE_TO = r'\\\g<0>'
EXTRACT = r'{\2}'


def find_routes(root):
    routes = []

    def find_node(node):
        if node.resource and node.resource.__class__.__name__ not in _doc_class_name:
            routes.append(node)

        for child in node.children:
            find_node(child)

    for route in root:
        find_node(route)

    return routes


def parse_path(path):
    subs, parameters = [], []
    for segment in path.strip('/').split('/'):
        matches = _FIELD_PATTERN.finditer(segment)
        if not matches:
            subs.append(segment)
            continue

        escaped = re.sub(ESCAPE, ESCAPE_TO, segment)
        subs.append(_FIELD_PATTERN.sub(EXTRACT, escaped))

        for field in matches:
            variable, converter, argstr = [field.group(name) for name in
                                           ('fname', 'cname', 'argstr')]

            if converter == 'int':
                args = [int(arg.strip()) for arg in argstr.split(',')]
                num_digits, minumum, maximum = args + [None] * (3 - len(args))
                schema = {
                    'type': 'integer',
                    'format': f'int{num_digits}' if num_digits else 'int32',
                }
                if minumum:
                    schema['minimum'] = minumum
                if maximum:
                    schema['maximum'] = maximum
            elif converter == 'uuid':
                schema = {
                    'type': 'string',
                    'format': 'uuid'
                }
            elif converter == 'dt':
                schema = {
                    'type': 'string',
                    'format': 'date-time',
                }
            else:
                # no converter specified or customized converters
                schema = {'type': 'string'}

            parameters.append({
                'name': variable,
                'in': 'path',
                'required': True,
                'schema': schema,
            })

    return f'/{"/".join(subs)}', parameters



class Config:
    def __init__(self):
        self.PATH = 'apidoc'  #: path for API document page
        self.TEMPLATE_FOLDER = 'templates'
        self.FILENAME = 'openapi.json'
        self.OPENAPI_VERSION = '3.0.2'  #: OpenAPI version

        self.TITLE = 'Service API Document'  #: service name
        self.VERSION = '0.1'  #: service version
        self.DOMAIN = None  #: service host domain


default_config = Config()
